from keras.models import Sequential
from keras.layers import Dense
from keras.layers import Dropout
from keras.layers import LSTM
from keras import losses
import numpy as np

# 最多8位二进制
BINARY_DIM = 8
#训练次数
epochs = 70
#生成的数据集大小
data_size = 50000

# 将整数表示成为binary_dim位的二进制数，高位用0补齐
def int_2_binary(number, binary_dim):
    binary_list = list(map(lambda x: int(x), bin(number)[2:]))
    number_dim = len(binary_list)
    result_list = [0]*(binary_dim-number_dim)+binary_list
    return result_list

# 将一个二进制数组转为整数
def binary2int(binary_array):
    out = 0
    for index, x in enumerate(reversed(binary_array)):
        out += x*pow(2, index)
    return out

# 将[0,2**BINARY_DIM)所有数表示成二进制
binary = np.array([int_2_binary(x, BINARY_DIM) for x in range(2**BINARY_DIM)])

#生成一个随机的两个数据和它的和
def batch_generation(largest_number):
    n1 = np.random.randint(0, largest_number)
    n2 = np.random.randint(0, largest_number)

    add = n1 +n2

    return np.append(binary[n1], binary[n2]),int_2_binary(add, BINARY_DIM+1)

# 样本的随机输入向量和输出向量
dataX = []
dataY = []
for i in range(data_size):
    inputX,inputY=batch_generation(2**BINARY_DIM)
    dataX.append(inputX)
    dataY.append(inputY)

# 重新特征X和目标变量Y数组，适应LSTM模型的输入和输出
X = np.reshape(dataX, (len(dataX), 2*BINARY_DIM, 1))
Y = np.array(dataY)

print(X.shape)
print(Y.shape)

# 定义LSTM模型
model = Sequential()
model.add(LSTM(256, input_shape=(X.shape[1], X.shape[2])))
model.add(Dropout(0.2))
model.add(Dense(Y.shape[1], activation='relu'))
model.compile(loss=losses.mean_squared_error, optimizer='adam')
model.summary()

model.fit(X, Y, epochs=epochs, batch_size=128)
mp = r'./LSTM_Operation.h5'
model.save(mp)

# use LSTM model to predict

# 样本的随机输入向量和输出向量
dataX = []
dataY = []
for i in range(1000):
    inputX,inputY=batch_generation(2**BINARY_DIM)
    dataX.append(inputX)
    dataY.append(inputY)

# 重新特征X和目标变量Y数组，适应LSTM模型的输入和输出
X = np.reshape(dataX, (len(dataX), 2*BINARY_DIM, 1))
Y = np.array(dataY)

test_acc = model.evaluate(X, Y)
print(str((1-test_acc)*100)+"%")

'''
test_acc = 0
for _ in range(1000):
    start = np.random.randint(0, len(dataX)-1)
    # print(dataX[start])
    number1 = dataX[start][0:BINARY_DIM]
    number2 = dataX[start][BINARY_DIM:]
    print('='*30)
    print('%s: %s'%(number1, binary2int(number1)))
    print('%s: %s'%(number2, binary2int(number2)))
    sample = np.reshape(X[start], (1, 2*BINARY_DIM, 1))
    predict = np.round(model.predict(sample), 0).astype(np.int32)[0]
    print('%s: %s'%(predict, binary2int(predict)))
    if binary2int(predict)==(binary2int(number1)+binary2int(number2)):
        test_acc=test_acc+1

print(test_acc/1000.0)

'''

